# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.monitor.abstract_hbase_monitor import AbstractHbaseMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.master.json_parser import JsonParser
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_metrics, \
    add_derived_jvm_metrics_jmx, add_metrics_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('MasterMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_ROLETYPE = 'MASTER'

# Section and key names in cloudera-monitor.properties
# Since 3.7
_SECTION = 'master'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_HBASE_MASTER_INFO_PORT = 'hbase_master_info_port'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.0
_SERVICE_VERSION = 'service_version'
# Added in CM 4.1
_LOG_DIR = 'log_dir'

class MasterMonitor(AbstractHbaseMonitor):

  def __init__(self, daemon, registry):
    AbstractHbaseMonitor.__init__(self,
                                  daemon,
                                  registry,
                                  _SECTION,
                                  _HBASE_MASTER_INFO_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._log_dir = None

  def __str__(self):
    return "MasterMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION,
              constants.SERVICE_VERSION_CDH3)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      host = confparse.get(_SECTION, _HOST)
      port = confparse.getint(_SECTION, _HBASE_MASTER_INFO_PORT)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(host, port)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._service_version = service_version
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect Master metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching Master metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version == constants.SERVICE_VERSION_CDH3:
      self._add_webserver_metrics_to_role_update_cdh3(update, metrics)
    elif self._service_version == constants.SERVICE_VERSION_CDH3U4X or \
         self._service_version == constants.SERVICE_VERSION_CDH4:
      self._add_webserver_metrics_to_role_update_cdh4(update, metrics)
    else:
      THROTTLED_LOG.warn("Unknown master version: %s" % (self._service_version))

  def _add_webserver_metrics_to_role_update_cdh3(self, update, metrics):
    add_derived_jvm_metrics_metrics(update, metrics)
    add_metrics_metrics(update, metrics, self._metric_accessors, _ROLETYPE,
                        'cdh3')

  def _add_webserver_metrics_to_role_update_cdh4(self, update, metrics):
    add_derived_jvm_metrics_metrics(update, metrics)
    add_metrics_metrics(update, metrics, self._metric_accessors, _ROLETYPE,
                        'cdh4')

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def get_subject_type(self):
    return "MASTER"
