# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import logging
import os
import simplejson as json
import shutil
import sys
import tarfile
import tempfile
import urlparse

import cmf.agent
import cmf.util

import diff
from parcel_id import ParcelId, VALID_EXT
import downloader

LOG = logging.getLogger(__name__)

PARCEL_CHECKSUM_HEADER = 'CM-Parcel-Checksum'

class ParcelCache(object):
  """
  Class that defines the parcel cache.

  The cache manages packed parcels. It decides when to download, where to
  store, and how to unpack parcels.
  """

  def __init__(self, path, repo, downloader, patch=diff.tarpatch):
    self.path = path
    self.repo = repo
    self.downloader = downloader
    self.tarpatch = patch
    LOG.info("Using %s for parcel cache" % (path))

  def _get_repo_parcels(self):
    parcels = set()
    for product in self.repo.products:
      for version in self.repo.products[product]:
        parcels.add(ParcelKey(product, version))
    return parcels

  def update(self, server_managed_parcels):
    """Update cache based on current list of server managed parcels. May
       download new or delete existing parcels.
    """
    managed_parcels = set()
    # for lookup from ParcelKey back to url
    parcel_url = dict()
    for url in server_managed_parcels:
      pid = ParcelId.url(url)
      key = ParcelKey(pid.product, pid.version)
      managed_parcels.add(key)
      parcel_url[repr(key)] = url

    available_parcels = self._get_repo_parcels()
    for unmanaged in (available_parcels - managed_parcels):
      self.delete_unmanaged(unmanaged.product, unmanaged.version)
    for newly_managed in (managed_parcels - available_parcels):
      url = parcel_url[repr(newly_managed)]
      # we do not skip if already exists because some retry mechanisms rely on
      # repeated, e.g. diff unpacking when base parcel not yet available
      self.queue_parcel_download(url, self.repo.path)

    # abort outdated downloads that are no longer managed
    for dl in self.downloader.get_current_downloads():
      url = dl[downloader.URL_FIELD]
      pid = ParcelId.url(url)
      key = ParcelKey(pid.product, pid.version)
      if key not in managed_parcels:
        self.downloader.abort(url)

  def prune_parcel_download_history(self, product, version):
    """Prune download history of urls that specify the parcel product and
       version.
    """
    def filter(url):
      return ParcelId.url(url).match(product, version)

    self.downloader.prune_history(filter)

  def delete_unmanaged(self, product, version):
    '''
    Delete parcel that is no longer managed by server, from both the cache
    and the repository. Any cached files with matching product and version
    are deleted.
    '''

    LOG.info("Deleting unmanaged parcel %s-%s" % (product, version))
    try:
      cache_files = os.listdir(self.path)
    except:
      LOG.exception("Exception while enumerating caches")
      return
    for cache_file in cache_files:
      filename, ext = os.path.splitext(cache_file)
      if ext is None or len(ext) < 1 or ext[1:] not in VALID_EXT:
        continue
      if ParcelId.filename(cache_file).match(product, version):
        try:
          os.remove(os.path.join(self.path, cache_file))
        except Exception:
          LOG.exception("Exception while removing cache %s" % (cache_file))

    # also delete from repo
    self.repo.delete(product, version)

  def queue_parcel_download(self, url, unpack_dir):
    """Queue a parcel download with auto unpacking upon completion. The url
       should download a file that ends with either .parcel or .diff. If a
       diff, the base parcel file is expected under the downloader path.
    """
    def cb(url, op):
      if op.state is downloader.ABORTED:
        # clean out the downloaded file and skip the unpacking
        LOG.info("Distribution aborted during download, skipping unpack.")
        if os.path.exists(op.path):
          os.remove(os.path)
        return True
      elif op.state is not downloader.DOWNLOADED:
        LOG.debug("Unable to unpack because download state is %s", str(op.state))
        return True

      pid = ParcelId.url(url)

      # If CM has sent a header with the checksum of the parcel we just
      # downloaded, calculate the checksum and raise an error if it does
      # not match.
      parcelHash = op.headers and op.headers.get(PARCEL_CHECKSUM_HEADER, None)
      if parcelHash:
        LOG.info('Checking checksum of parcel %s...', pid.get_filename())
        digest = cmf.util.sha1(op.path)
        if parcelHash.lower() != digest:
          LOG.warn('Parcel %s did not match checksum provided by server: %s != %s.',
              pid.get_filename(), parcelHash, digest)
          raise Exception, 'Checksum mismatch'

      unpacked_path = os.path.join(unpack_dir, pid.get_parcel_dir())
      if os.path.exists(unpacked_path):
        LOG.info("%s already exists, skipping unpacking." % (unpacked_path))
        return True
      try:
        LOG.info("Unpacking %s into %s" % (op.path, unpack_dir))
        if pid.is_diff:
          base_parcel = os.path.join(self.path, pid.get_base_parcel_filename())
          if not os.path.exists(base_parcel):
            # base parcel not yet distributed. retry on next hb->queue->cb.
            return False
          else:
            self.tarpatch(base_parcel, op.path, unpack_dir, self.path)
        else:
          unpack_parcel(op.path, unpack_dir, self.path)
      except Exception, e:
        LOG.exception("Error while unpacking %s" % (op.path))
        raise e # makes download state EXCEPTION and terminal
      return True

    def unpack_parcel(src, dst, tmp):
      # unpack into dir named as <product>-<version>
      if not os.path.exists(dst):
        LOG.info("Making unpack dir %s" % (dst))
        os.makedirs(dst)
      if not os.path.exists(src):
        raise Exception("Src file %s does not exist" % (src))

      tar = tarfile.open(src, 'r')
      try:
        if hasattr(tar, 'extractall'):
          # New in version 2.5
          tar.extractall(tmp)
        else:
          for member in tar.getmembers():
            tar.extract(member, tmp)
      finally:
        tar.close()
      dirname = ParcelId.url(src).get_parcel_dir()
      shutil.move(os.path.join(tmp, dirname), os.path.join(dst, dirname))

    # ParcelId helps validate url
    ParcelId.url(url)
    self.downloader.queue_download(url, callback=cb, proceed=True)

class ParcelKey(object):
  def __init__(self, product, version):
    self.product = product
    self.version = version

  def __eq__(self, other):
    return self.product == other.product and self.version == other.version

  def __hash__(self):
    return hash(self.product) ^ hash(self.version)

  def __str__(self):
    return "%s-%s" % (self.product, self.version)
