#!/bin/bash
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

# defines a set of helper functions that can be used by Cloudera
# manager service scripts and exports some common environment
# variables.
#

# Attempts to locate java home, prints an error and exits if no
# java can be found.
locate_java_home() {
# attempt to find java
  if [ -z "$JAVA_HOME" ]; then
    for candidate in \
      /usr/lib/j2sdk1.6-sun \
      /usr/lib/jvm/java-6-sun \
      /usr/lib/jvm/java-1.6.0-sun-1.6.0.* \
      /usr/lib/jvm/java-1.6.0-sun-1.6.0.*/jre/ \
      /usr/lib/jvm/j2sdk1.6-oracle \
      /usr/lib/jvm/j2sdk1.6-oracle/jre \
      /usr/java/jdk1.6* \
      /usr/java/jre1.6* \
      /Library/Java/Home \
      /usr/java/default \
      /usr/lib/jvm/default-java \
      /usr/lib/jvm/java-openjdk \
      /usr/lib/jvm/jre-openjdk \
      /usr/lib/jvm/java-1.6.0-openjdk-1.6.* \
      /usr/lib/jvm/jre-1.6.0-openjdk* ; do
      if [ -e $candidate/bin/java ]; then
        export JAVA_HOME=$candidate
        break
      fi
    done
    # if we didn't set it
    if [ -z "$JAVA_HOME" ]; then
      cat 1>&2 <<EOF
+======================================================================+
|      Error: JAVA_HOME is not set and Java could not be found         |
+----------------------------------------------------------------------+
| Please download the latest Oracle JDK from the Oracle Java web site  |
|  > http://www.oracle.com/technetwork/java/javase/index.html <        |
|                                                                      |
| Cloudera Manager requires Java 1.6 or later.                         |
| NOTE: This script will find Oracle Java whether you install using    |
|       the binary or the RPM based installer.                         |
+======================================================================+
EOF
      exit 1
    fi
  fi
}

# Source the parcel environment scripts passed by the agent
source_parcel_environment() {
  if [ ! -z "$SCM_DEFINES_SCRIPTS" ]; then
    # Narrow IFS to newline only to allow embedded spaces
    OLD_IFS=$IFS
    IFS=$':'
    SCRIPT_ARRAY=($SCM_DEFINES_SCRIPTS)
    IFS=$OLD_IFS
    for SCRIPT in ${SCRIPT_ARRAY[@]}; do
      . "$SCRIPT"
    done
  fi
}

# sets the default HADOOP_CLASSPATH to include cloudera log4j appender.
# The function will append to an existing HADOOP_CLASSPATH variable if it
# is defined. MGMT_HOME must be defined, otherwise the function prints an
# error message and exits.
set_hadoop_classpath() {
  set_classpath_in_var HADOOP_CLASSPATH
}

# sets the default HBASE_CLASSPATH to include cloudera log4j appender.
# The function appends to an existing HADOOP_CLASSPATH variable if it
# is defined. MGMT_HOME must be defined, otherwise the function prints an
# error message and exits.
set_hbase_classpath() {
  set_classpath_in_var HBASE_CLASSPATH
}

# sets the default ZOOKEEPER_CLASSPATH to include cloudera log4j appender.
# The function appends to an existing ZOOKEEPER_CLASSPATH variable if it
# is defined. MGMT_HOME must be defined, otherwise the function prints an
# error message and exits.
set_zookeeper_classpath() {
  set_classpath_in_var ZOOKEEPER_CLASSPATH
}

# sets the default FLUME_CLASSPATH to include cloudera log4j appender.
# The function appends to an existing FLUME_CLASSPATH variable if it
# is defined. MGMT_HOME must be defined, otherwise the function prints an
# error message and exits.
set_flume_classpath() {
  set_classpath_in_var FLUME_CLASSPATH
}

# sets hive's AUX_CLASSPATH to include cloudera governor plugin jar.
# MGMT_HOME must be defined, otherwise the function prints an
# error message and exits.
set_hive_classpath() {
  set_classpath_in_var AUX_CLASSPATH 
}

# sets the classpath variable passed as a parameter to include all of cloudera
# default classes.
#
# If no parameter is passed the function prints an error message and exits.
#
# Currently this function includes only one jar, the Cloudera log4j
# appender in the default class path. Note that MGMT_HOME must be defined,
# if it is not the function prints an error and exits.
#
# param $1 - The name of the classpath variable to assign to. The function
# appends all the extra jars to this varible.
set_classpath_in_var() {
  if [ -z $1 ]; then
    echo "Must call with the name of variable to assign."
    exit 1
  fi
  if [ "$MGMT_HOME" ]; then
    # Add all plugin jars to the classpath.
    ADD_TO_CP=`find $MGMT_HOME/lib/plugins -name '*.jar' | tr "\n" ":"`
    ADD_TO_CP=${ADD_TO_CP%?}
    eval OLD_VALUE=\$$1
    if [ -z $OLD_VALUE ]; then
      export $1="$ADD_TO_CP"
    else
      export $1="$ADD_TO_CP:$OLD_VALUE"
    fi
  else
    echo "MGMT_HOME must be set."
    exit 1
  fi
}

acquire_kerberos_tgt() {
  if [ -z $1 ]; then
    echo "Must call with the name of keytab file."
    exit 1
  fi

  # Acquire Kerberos tgt (ticket-granting ticket) if the server provided the
  # principal (in which case the keytab should be non-zero).
  #
  # Note that we cache it in the current directory so that it will be isolated to
  # this hadoop command.
  if [ -n "$SCM_KERBEROS_PRINCIPAL" ]; then
    if [ -d /usr/kerberos/bin ]; then
      export PATH=/usr/kerberos/bin:$PATH
    fi
    which kinit
    if [ $? -ne 0 ]; then
      echo "kinit does not exist on the host."
      exit 1
    fi

    export KRB5CCNAME=$CONF_DIR/krb5cc_$(id -u)
    echo "using $SCM_KERBEROS_PRINCIPAL as Kerberos principal"
    echo "using $KRB5CCNAME as Kerberos ticket cache"
    kinit -c $KRB5CCNAME -kt $CONF_DIR/$1 $SCM_KERBEROS_PRINCIPAL
    if [ $? -ne 0 ]; then
      echo "kinit was not successful."
      exit 1
    fi
    # This is work-around for a bug in kerberos >= 1.8 that prevents java
    # programs from reading from the ticket cache. It's harmless to do it
    # unconditionally - as long as we sleep first, in case kerberos is
    # configured to prevent ticket renewal. If the two kinit calls are
    # too close together, the -R can succeed when it shouldn't, and end
    # up expiring the ticket.
    sleep 1
    kinit -R
  fi
}

# SHOULD NOT BE USED DIRECTLY OUTSIDE OF THIS FILE.
# Helper function to report a CM specific status code. Expects the CM status
# code to report and the exit code to exit with.
__cm_report_status_and_exit_with_code() {
  if [ $# -ne 2 ]; then
    echo "expected 2 arguments - CM status code and exit code" 1>&2
    exit 1
  fi

  # Note that CM_STATUS_CODES is injected into the env by the agent.
  for code in $CM_STATUS_CODES; do
    if [ "$code" == "$1" ]; then
      # Output the error in the format
      # "CM_STATUS_CODE=<value>" as the last line of the stderr stream. Note
      # that this format and writing to stderr is important since the agent will
      # otherwise not be able to extract this error code and pass it along to
      # CM.
      echo CM_STATUS_CODE=$code 1>&2
      exit $2
    fi
  done

  echo "Unexpected CM error code: $1" 1>&2
  exit 1
}

# Reports a CM specific status code and exits with error i.e. with exit code 1.
cm_error_and_exit() {
  # Suppress tracing for this function so that the below script does not show up
  # in the error logs shown to users.
  #
  # Not suppressing the tracing would require making the agent's parsing code
  # more complex and sub-optimal (since more text will need to be searched).
  # Additionally, there is a loop in this code (in the helper function being
  # invoked below) which can cause a lot of text to be output to the error log
  # as the number of enum values increases.
  set +xv

  if [ $# -ne 1 ]; then
    echo "expected 1 argument - CM status code to report" 1>&2
    exit 1
  fi

  __cm_report_status_and_exit_with_code $1 1
}

# Report a CM specific status code and exit successfully i.e. with exit code 0.
cm_success_and_exit() {
  # Suppress tracing for this function so that the below script does not show up
  # in the error logs shown to users.
  #
  # Not suppressing the tracing would require making the agent's parsing code
  # more complex and sub-optimal (since more text will need to be searched).
  # Additionally, there is a loop in this code (in the helper function being
  # invoked below) which can cause a lot of text to be output to the error log
  # as the number of enum values increases.
  set +xv

  if [ $# -ne 1 ]; then
    echo "expected 1 argument - CM status code to report" 1>&2
    exit 1
  fi

  __cm_report_status_and_exit_with_code $1 0
}

copy_hive_default() {
# CDH3 Hive expects the hive-default.xml to be in the path
  if [[ -z "$CDH_VERSION" ]]; then
    echo "ERROR: CDH version must be provided"
    exit 1
  elif [ $# -ne 1 ]; then
    echo "ERROR: expected 1 argument - destination dir for $HIVE_DEFAULT_XML"
    exit 1
  elif [[ "$CDH_VERSION" == 3 ]]; then
    echo "HIVE_DEFAULT_XML=$HIVE_DEFAULT_XML"
    echo "Copying hive-default.xml to $1"
    if !(cp "$HIVE_DEFAULT_XML" "$1"); then
      echo "ERROR: failed to copy hive-default.xml"
      exit 1
    fi
  fi
}

# Enable tracing.
set -x
