#!/usr/bin/env python
# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
import codecs
import logging
import os
import pwd
import grp
import shutil
from stat import S_ISDIR
import sys
from string import Template
import subprocess
try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO
from zipfile import ZipFile

try:
  from hashlib import sha1 as _sha1
except ImportError:
  from sha import new as _sha1

import urllib2

LOG = logging.getLogger(__name__)

def sha1(path):
  """
  Calculate the SHA-1 checksum of the file at the given path. Returns the
  lower-case hex string representing the checksum.
  """
  digest = _sha1()
  f = open(path, 'r')
  try:
    buf = f.read(4096)
    while buf:
      digest.update(buf)
      buf = f.read(4096)
  finally:
    f.close()
  return digest.hexdigest().lower()

# urlopen_with_timeout.
# The timeout parameter to urlopen was introduced in Python 2.6.
# To workaround it in older versions of python, we copy, with
# minor modification, httplib.HTTPConnection, and hook it all
# up.
if sys.version_info < (2, 6):
  import httplib
  import socket

  def urlopen_with_timeout(url, data=None, timeout=None):
    # Create these two helper classes fresh each time, since
    # timeout needs to be in the closure.

    class TimeoutHTTPConnection(httplib.HTTPConnection):
      def connect(self):
        """Connect to the host and port specified in __init__."""
        msg = "getaddrinfo returns an empty list"
        for res in socket.getaddrinfo(self.host, self.port, 0,
                        socket.SOCK_STREAM):
          af, socktype, proto, canonname, sa = res
          try:
            self.sock = socket.socket(af, socktype, proto)
            if timeout is not None:
              self.sock.settimeout(timeout)
            if self.debuglevel > 0:
              print "connect: (%s, %s)" % (self.host, self.port)
            self.sock.connect(sa)
          except socket.error, msg:
            if self.debuglevel > 0:
              print 'connect fail:', (self.host, self.port)
            if self.sock:
              self.sock.close()
            self.sock = None
            continue
          break
        if not self.sock:
          raise socket.error, msg

    class TimeoutHTTPHandler(urllib2.HTTPHandler):
      http_request = urllib2.AbstractHTTPHandler.do_request_
      def http_open(self, req):
        return self.do_open(TimeoutHTTPConnection, req)

    opener = urllib2.build_opener(TimeoutHTTPHandler)
    return opener.open(url, data)

else:
  urlopen_with_timeout = urllib2.urlopen


def unzip(zip_data, target_directory, uid=None, gid=None, namelist=None):
  """
  Unzips a zipfile into target_directory, with mode 600 (files) or 700
  (directories). User and group ownership is set for newly created files
  and directories.

  python2.4 has very rudimentary zipfile support;
  in newer versions this is just ZipFile.extract_all.

  namelist parameter takes a list of names to unzip. These must exactly
  match the names, including directories, that appear in the zip file.
  """
  count = 0
  dir_count = 0
  umask = os.umask(0077)
  try:
    zip_file = None
    if zip_data != "":
      zip_file = ZipFile(StringIO(zip_data), "r")

    if namelist == None:
      if zip_file != None:
        namelist = zip_file.namelist()
      else:
        namelist = []

    for f in namelist:
      try:
        zip_file.getinfo(f)
      except KeyError:
        LOG.warn("Not extracting'%s' because it's not in the archive." %f)
        continue
      target = os.path.normpath(os.path.join(target_directory, f))
      if target_directory != os.path.commonprefix([target_directory, target]):
        LOG.error("Not extracting '%s' because it wishes to escape the target directory." % f)
        continue
      d = os.path.dirname(target)
      dirs_to_create = []
      while not os.path.exists(d):
        assert target_directory == os.path.commonprefix([d, target_directory])
        dirs_to_create.append(d)
        d = os.path.dirname(d)
      for d in reversed(dirs_to_create):
        os.mkdir(d)
        dir_count += 1
        if uid != None:
          os.chown(d, uid, gid)
      exist = os.path.exists(target)
      out = file(target, "w")
      out.write(zip_file.read(f))
      out.close()
      if uid != None and not exist:
        os.chown(target, uid, gid)
      count += 1
    LOG.info("Extracted %d files and %d dirs to %s." % (count, dir_count, target_directory))
  finally:
    os.umask(umask)

def tail(file_or_filename, length):
  """
  Reads last (at most) length of the given file or filename.
  We take some pains to make sure that the first line is after
  a newline, but we don't worry about partial lines at
  the end.
  """
  assert length >= 0
  if hasattr(file_or_filename, "seek"):
    f = codecs.EncodedFile(file_or_filename, "ascii", "ascii", "ignore")
  else:
    f = codecs.open(file_or_filename, 'r', "ascii", "ignore")

  f.seek(0, 2) # 2==os.SEEK_END, which isn't available on py2.4
  f.seek(-1*(min(length, f.tell())), 2)
  now = f.tell()
  t = f.read(length)
  # t is, at this point, the last 'length' of file
  # (or the entire if file, if it's too small)
  if now == 0:
    return t
  else:
    # Ignore a potential partial line
    return t[t.find("\n") + 1:]

def user_group_to_ids(user, group):
  """Given username and group name, returns ids thereof."""
  return pwd.getpwnam(user).pw_uid, grp.getgrnam(group).gr_gid

def dumpstacks(signal__unused, frame__unused):
  import threading
  import sys
  import traceback
  id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
  code = []
  for threadId, stack in sys._current_frames().items():
    code.append("\n# Thread: %s(%d)" % (id2name[threadId], threadId))
    for filename, lineno, name, line in traceback.extract_stack(stack):
      code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
      if line:
        code.append("  %s" % (line.strip()))
    print "\n".join(code)

def register_sigquit_stack_dump():
  import signal
  signal.signal(signal.SIGQUIT, dumpstacks)

def delete_quietly(filepath):
  if filepath is not None:
    try:
      os.remove(filepath)
    except:
      pass

def close_quietly(fd):
  if fd is not None:
    try:
      fd.close()
    except:
      pass

def deploy_special_file(special_file_info, src_dir, env_vars, change_owner):
  '''
  This function applies the instructions given in special_file_info
  by changing the location, owner and permissions of a file present in
  src_dir. Owner and permissions are changed only if change_owner is True.
  All env_vars starting with '$' are replaced by their values in env_vars dict.
  '''

  orig_file = os.path.join(src_dir, special_file_info["name"])
  if change_owner:
    info_uid, info_gid = \
      user_group_to_ids(special_file_info["user"], special_file_info["group"])
    os.chown(orig_file, info_uid, info_gid)
  os.chmod(orig_file, special_file_info["permissions"])

  deploy_dir = Template(special_file_info["deploy_dir"]).substitute(env_vars)
  dest_file = deploy_dir + "/" + special_file_info["name"]
  shutil.move(orig_file, dest_file)
  LOG.info("Moved %s to %s." % (orig_file, dest_file))

def bytes_to_hex(b):
  return ''.join( [ "%02x" % ord( x ) for x in b ] )

def validate_file_owner_from_stat(stat, valid_owners):
  """
  Checks if the owner of the file the stat belongs to is in the list of
  valid owners.
  """
  uid = stat.st_uid
  try:
    user = pwd.getpwuid(uid)[0]
  except KeyError:
    raise Exception("Unable to find username with uid %d" % (uid,))

  if user not in valid_owners:
    raise Exception("Invalid file owner: %s" % (user,))

def validate_file_owner(path, valid_owners):
  """
  Checks if the owner of the file specified by the path is in the list of
  valid owners.
  """
  stat = os.stat(path)
  validate_file_owner_from_stat(stat, valid_owners)

def validate_and_open_fd(path, parent_path=None, no_links=True, valid_owners=None):
  """
  Returns a file descriptor after performing some validations on the path and file.
  The validations are:
    - The file exists
    - Whether links should be followed (no symlinks, no files with more than
      one hard link).
    - Whether to restrict the path to be within a specified parent directory
    - Restrict the owners of the file to set of valid owners

  If any of these validations fail an exception is thrown.

  The file may also be a directory.
  """
  # Get rid of any up level references ("..")
  path = os.path.normpath(path)
  if parent_path is not None:
    parent_path = os.path.normpath(parent_path)

  # Check file falls within the process directory if needed
  if parent_path is not None:
    if not path.startswith(parent_path + '/'):
      raise Exception("Path is not within process directory: %s" % (path,))

  flags = os.O_RDONLY
  if no_links or parent_path is not None:
    # Don't follow symlinks
    flags |= os.O_NOFOLLOW

  fd = os.open(path, flags)
  try:
    stat = os.fstat(fd)

    # If we don't want to follow links, we want to prevent hardlinks. Since
    # anyone can create a hardlink to any file (ubuntu being one exception)
    # the only thing we can do is disallow giving access to files that have
    # more than one hardlink. Note that directories have multiple hardlinks
    # by their nature and can't be hardlinked by users, so skip this test
    # for them.
    if (no_links or parent_path is not None) and not S_ISDIR(stat.st_mode):
      if stat.st_nlink > 1:
        raise Exception("Path has more than one hardlink: %s" % (path,))

    # Validate owner if needed
    if valid_owners is not None:
      validate_file_owner_from_stat(stat, valid_owners)
  except Exception, e:
    os.close(fd)
    raise e

  return fd

def validate_and_open(path, parent_path=None, no_links=True, valid_owners=None):
  """
  Returns a file object after performing some validations on the path and file.

  See {@link validate_and_open_fd} for details on the validations performed.

  If any of these validations fail an exception is thrown.

  While validate_and_open_fd can handle a directory, this method can only
  handle files as python file objects cannot wrap a directory fd.
  """
  return os.fdopen(validate_and_open_fd(path, parent_path, no_links, valid_owners))

def source(path, command, caller_env):
  """
  Source a shell script in a subprocess and extract its environment and return
  it as a dict. For our specific purposes we need to support executing a command
  after sourcing the script to correctly populate the environment.
  """

  pipe = subprocess.Popen(['/bin/bash', '-c', ". %s; %s; env" % (path, command)],
                          stdout=subprocess.PIPE, env=caller_env)
  data = pipe.communicate()[0]

  return dict((line.split("=", 1) for line in data.splitlines()))
