// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'knockout',
  'underscore',
  'cloudera/Util'
], function(ko, _, Util) {
  // This function is bound to every filter object we get from the server.
  var has = function() {
    var self = this;
    // All of the attributes must be found to return true.
    return _.all(arguments, function(attr) {
      return self.hasOwnProperty(attr);
    });
  };

  // This function is bound to every filter object we get from the server.
  var hasnt = function() {
    var self = this;
    // None of the attributes must be found to return true.
    return !_.all(arguments, function(attr) {
      return self.hasOwnProperty(attr);
    });
  };

  // This function is bound to a filter object we get from the server if the
  // filter has an attribute 'valueUri'; this means that it is a
  // auto-complete filter and should query the server for values to help
  // the user.
  var executeQuery = function(query) {
    // Do nothing if the user hasn't entered anything yet.
    if (!query.term) {
      return;
    }
    var args = {
      q: query.term
    };
    $.getJSON(this.valueUri, args, function(data) {
      // The server is sending back a simple list of strings. We have
      // to decompose it for select2 in a particular way.
      var results = _.map(data, function(result) {
        return {
          id: result,
          text: result
        };
      });
      query.callback({
        results: results
      });
    });
  };

  // This is only used locally by the FilterModel::values property. If this
  // ends up being generally useful we can extract it to its own file.
  ko.extenders.deepArrayEquals = function(target) {
    return ko.computed({
      // Always return the target value.
      read: target,

      write: function(newValue) {
        var current = target();

        // Only set if the two values are not already equal.
        if (_.isArray(newValue) && !_.isEqual(newValue, current)) {
          target(newValue);
        }
      }
    });
  };

  var FilterModel = function(filterViewModel) {
    var self = this;
    self.filter = ko.observable();
    self.compareType = ko.observable();
    // So... this is really stupid. The select2 control and the
    // selectedOptions binding don't play well together. When used together
    // they create an infinite loop of updates on initial construction of
    // the control. Apparently KO doesn't recognize that one empty array
    // is equal to another empty array; when initially constructed the
    // select2 control sets the <select> element to empty, which triggers
    // the observableArray to notify observers, which makes the ko.select2
    // binding trigger a change event on select2, which makes it set the
    // <select> element to empty... on and on it goes. I have used an
    // extender on this observableArray to only notify when a change
    // actually changes the array. It also double-checks that the value
    // assigned to the observableArray is an array.
    self.values = ko.observableArray().extend({ deepArrayEquals: null });
    // A filter is active if the user is currently editing it.
    self.active = ko.observable();

    // The following computed observables make templating easier.
    self.filterName = ko.computed({
      read: function() {
        var filter = self.filter();
        return filter ? filter.propertyName : null;
      },

      write: function(newValue) {
        if (!filterViewModel) {
          return;
        }
        // Find the filter in the existing list of filters.
        var filter = _.find(filterViewModel.filters(), function(filter) {
          return filter.propertyName === newValue;
        });
        self.filter(filter);
      }
    });

    self.compareTypeName = ko.computed({
      read: function() {
        var compareType = self.compareType();
        return compareType ? compareType.name : null;
      },

      write: function(newValue) {
        if (!self.filter()) {
          return;
        }
        var compareType = _.find(self.filter().compareTypes, function(compareType) {
          return compareType.name === newValue;
        });
        self.compareType(compareType);
      }
    });

    self.value = ko.computed({
      read: function() {
        return self.values().join(',');
      },
      write: function(newValue) {
        if (newValue) {
          var values = newValue.split(',');
          self.values(values);
        } else {
          self.values([]);
        }
      }
    });

    self.filter.subscribe(function(newFilter) {
      if (!newFilter || !newFilter.compareTypes ||
          newFilter.compareTypes.length !== 1) {
        return;
      }
      var compareType = newFilter.compareTypes[0];
      self.compareType(compareType);
    });

    // A FilterModel instance is complete if it has a filter, a compare type,
    // and at least one value in the values array.
    self.complete = ko.computed(function() {
      return self.filter() && self.compareType() &&
        self.values() && self.values().length;
    });
    
    self.activate = function(data, e) {
      if (!self.active() && !$(e.target).is('.btn')) {
        self.active(true);
      }
    };

    self.compareTo = function(to) {
      return to && self.filter() === to.filter()
          && self.compareType() === to.compareType()
          && _.isEqual(self.values().sort(), to.values().sort());
    };
  },
  FilterViewModel = function(eventSearch) {
    var self = this;
    self.loading = eventSearch.loading;
    self.filters = eventSearch.filters;
    self.appliedFilters = eventSearch.appliedFilters;
    self.activeFilter = eventSearch.activeFilter;
    self.removeFilter = _.bind(eventSearch.removeFilter, eventSearch);
    self.findFilter = _.bind(eventSearch.findFilter, eventSearch);

    self.appendFilter = function(filter, compareType, value) {
      var filterModel = new FilterModel(this),
        aFilters = self.appliedFilters();
      if (filter) {
        filterModel.filter(filter);
        if (compareType) {
          if (_.isString(compareType)) {
            compareType = eventSearch.compareTypeFromName(filter,
                compareType);
            if (!compareType) {
              return false;
            }
          }
          filterModel.compareType(compareType);
          if (value) {
            filterModel.value(value);
          }
        }
      }
      return eventSearch.appendFilter(filterModel);
    };

    self.deactivateFilters = function() {
      var active = self.activeFilter();
      if (active) {
        active.active(false);
      }
    };

    self.addAnother = function() {
      self.deactivateFilters();
      self.appendFilter();
    };

    // Called by select2 to let the user input a freeform choice.
    self.createSearchChoice = function(term, matches) {
      return {
        id: term,
        text: term
      };
    };

    // This is used to provide an initial selection to the select2 control.
    // select2 expects a list of objects of the form {id:, text:} when it
    // first parses the values out. Since the input[text] that this is attached
    // to can't store objects, it stores a comma-delimited list of values.
    // These values need to be converted to the select2 format during the data-
    // bind phase; hence this method.
    self.fromCommaDelimitedValues = function(element, callback) {
      var value = $(element).val(), values = [];
      if (value) {
        values = value.split(',');
      }
      var result = _.map(values, function(value) {
        return {
          id: value,
          text: value
        };
      });
      callback(result);
    };

    if (eventSearch.exportEvents) {
      self.exportEvents = eventSearch.exportEvents;
    }
  },
  EventSearch = function(filterContainer, eventList, filterPath,
      staticFilters, dataSource, exportEvents) {

    this.error = ko.observable();

    this.filterContainer = filterContainer;
    this.filters = ko.observableArray();
    this.staticFilters = staticFilters || [];
    this.appliedFilters = ko.observableArray();
    this.filterPath = filterPath;
    this.activeFilter = ko.observable();

    this.eventList = eventList;
    this.events = ko.observableArray();
    this.loading = ko.observable(false);
    this.dataSource = dataSource;
    this.exportEvents = exportEvents;
    this.viewModel = new FilterViewModel(this);

    // Return true if there is a filter with the given name.
    var hasFilterNamed = function(name) {
      return _.some(this.filters(), function(filter) {
        return filter.propertyName === name;
      });
    };

    ko.applyBindings(this.viewModel, filterContainer);

    ko.applyBindings({
      loading : this.loading,
      error : this.error,
      events : this.events,
      hasFilterNamed: _.bind(hasFilterNamed, this)
    }, eventList);
  };

  // Bound to an instance of EventSearch in the method appendFilter below.
  var onFilterModelActiveChanged = function(filterModel, newValue) {
    var currentlyActive = this.activeFilter();
    // Is a filter becoming active?
    if (newValue) {
      // Deactivate current filter if not same as filterModel.
      if (currentlyActive && currentlyActive !== filterModel) {
        currentlyActive.active(false);
      }
      // This is the new active filter.
      this.activeFilter(filterModel);
    } else {
      // Is the currently active the one that isn't active anymore?
      if (currentlyActive === filterModel) {
        if (!currentlyActive.complete()) {
          this.removeFilter(currentlyActive);
        }
        this.activeFilter(null);
      }
      // If this filter is no longer active but is complete, fire
      // the appliedfilters.change event.
      if (filterModel.complete()) {
        $.publish('eventsearch.appliedfilters.changed');
      }
    }
  };

  EventSearch.prototype = {
    compareTypeFromName : function(filter, name) {
      return _.find(filter.compareTypes, function(compareType) {
        return compareType.name === name;
      });
    },
    getTail : function() {
      return this.length() ? this.events()[this.length() - 1] : null;
    },
    getHead : function() {
      return this.length() ? this.events()[0] : null;
    },
    length : function() {
      return this.events().length;
    },
    splice : function() {
      return this.events.splice
          .apply(this.events, $.makeArray(arguments));
    },
    deactivateAllFilters : function() {
      var active = this.activeFilter();
      if (active) {
        active.active(false);
      }
    },
    prepend : function(event) {
      var result;
      if (_.isArray(event)) {
        this.events(this.length() ? event.concat(this.events()) : event);
        result = this.length();
      } else {
        result = this.events.unshift(event);
      }
      return result;
    },
    append : function(event) {
      var result;
      if (_.isArray(event)) {
        this.events(this.length() ? this.events().concat(event) : event);
        result = this.length();
      } else {
        result = this.events.push(event);
      }
      return result;
    },
    loadFilters : function(callback) {
      var eventSearch = this;
      $.getJSON(this.filterPath, _.bind(this.loadFiltersHandler, this))
        .complete(callback);
    },
    loadFiltersHandler: function(data) {
      var response = Util.filterError(data);
      this.clearFilters();
      this.registerAllFilters(response.data);
    },
    retryHandlerAfterTime: function(clear, handler, timeInMs) {
      var self = this;
      setTimeout(function() {
        self.dataSource(self, clear, handler);
      }, timeInMs);
    },
    makeLoadDataHandler: function(clear, callback) {
      var self = this;
      return function(err) {
        self.loading(false);

        if (err) {
          self.error(err);
        }

        if (callback && callback(err)) {
          var boundHandler = _.bind(self.makeLoadDataHandler, self);
          self.retryHandlerAfterTime(clear, boundHandler, 2000);
        }
      };
    },
    loadData : function(clear, callback) {
      if (_.isFunction(clear)) {
        callback = clear;
        clear = undefined;
      }

      this.error(undefined);
      this.loading(true);

      var handler = this.makeLoadDataHandler(clear, callback);
      this.dataSource(this, clear, handler);
    },
    getAppliedFilters : function() {
      // Return a set of applied filters starting with a copy of any
      // static filters.
      var results = this.staticFilters.slice(0);
      _.each(this.appliedFilters(), function(filter) {
        // Make sure the filter is complete and that its compareType
        // has a name before returning it.
        if (filter.complete() && filter.compareType().name) {
          results.push({
            propertyName: filter.filter().propertyName,
            compareType: filter.compareType().name,
            values: filter.values()
          });
        }
      });
      return results;
    },
    appendFilter: function(filterModel) {
      // If there are any duplicates, stop here.
      var areEqual = function(filter) {
        return filter.compareTo(filterModel);
      };
      if (_.any(this.appliedFilters(), areEqual)) {
        return false;
      }
      // Subscribe to its active observable.
      var activeChanged = _.bind(onFilterModelActiveChanged, this, filterModel);
      filterModel.active.subscribe(activeChanged);
      this.appliedFilters.push(filterModel);
      // If it's not complete, then it is the new active filter.
      filterModel.active(!filterModel.complete());
      return true;
    },
    removeFilter: function(removeMe) {
      // I'm not using _.filter here because I don't want to completely
      // erase the array. Knockout observable arrays take a bit of care.
      this.appliedFilters.remove(removeMe);
      // Are we removing the active filter? If so, deactivate it.
      if (removeMe && removeMe.compareTo(this.activeFilter())) {
        this.activeFilter(null);
      }
      $.publish('eventsearch.appliedfilters.changed');
      return false;
    },
    applyFilter : function(filter, compareType, value) {
      return this.viewModel.appendFilter(filter, compareType, value);
    },
    clear : function() {
      this.events.splice(0);
    },
    findFilter : function(name) {
      return _.find(this.filters(), function(filter) {
        return filter.propertyName === name;
      });
    },
    registerFilter : function(filter) {
      if (_.find(this.staticFilters, function(staticFilter) {
        return staticFilter.propertyName === filter.propertyName;
      })) {
        return;
      }
      // Assign some utility methods to the filter object for use in
      // the KO bindings in the template.
      filter.has = has;
      filter.hasnt = hasnt;
      if (filter.valueUri) {
        filter.executeQuery = _.bind(executeQuery, filter);
      }

      this.filters.push(filter);
    },
    registerAllFilters : function(filters) {
      _.each(filters, this.registerFilter, this);
    },
    clearFilters : function() {
      this.filters.splice(0);
    }
  };

  // Expose FilterModel and FilterViewModel for testing.
  EventSearch.FilterModel = FilterModel;
  EventSearch.FilterViewModel = FilterViewModel;

  return EventSearch;
});
