// (c) Copyright 2012 Cloudera, Inc. All rights reserved.
define([
  "underscore",
  "cloudera/Util"
], function (_, Util)  {

  // TODO: Make this configurable, either dynamically from the controller or
  // from a config value exposed as config on clouderaManager object.
  var BUCKET_SIZE = 4;

  var padTsResponse = function(timeSeriesResponse, queryParams) {
    if (_.isEmpty(timeSeriesResponse.timeSeries)) {
      timeSeriesResponse.timeSeries = [ {
        data: [ {
          x: queryParams.startTime,
          y: 0,
          type: "fake"
        }, {
          x: queryParams.endTime,
          y: 0,
          type: "fake"
        } ],
        metadata: {
          label: "",
          attributes: {}
        }
      } ];
    } else {
      _.each(timeSeriesResponse.timeSeries, function(ts) {
        if (_.isEmpty(ts.data)) {
          ts.data = [ {
            x: queryParams.startTime,
            y: 0,
            type: "fake"
          }, {
            x: queryParams.endTime,
            y: 0,
            type: "fake"
          } ];
        }
      });
    }
  };

  /**
   * Bulk fetches a set of tsquery as one request.
   *
   * options = {
   *   container:      (required) "selector or element of the container DOM object",
   *   plotContainers: (required) a list of PlotContainer objects.
   *   updateRecent:   (optional) a flag to indicate if we should update the recent queries.
   *   enableFeedbackErrors: (optional) true|false Whether to fire error messages, defaults true.
   *   enableFeedbackWarnings: (optional) true|false Whether to fire warning messages, defaults true.
   * };
   */
  function BulkTsqueryFetcher(options) {
    var self = this, $container = $(options.container);

    var enableFeedbackErrors = (options.enableFeedbackErrors === false) ? false : true;
    var enableFeedbackWarnings = (options.enableFeedbackWarnings === false) ? false : true;

    // This id is send with the error and warning messages to indicate to help
    // the receiver scope which thing is sending it error or warning messages.
    var feedbackScopeId = _.uniqueId('query-fetcher-');

    var createMessageHandler = function(messageType, eventType) {
      return function(data) {
        var messages = [];
        _.each(data, function(timeSeriesResponse) {
          _.each(timeSeriesResponse[messageType], function(text) {
            messages.push(text);
          });
        });
        $.publish(eventType, [messages, feedbackScopeId]);
      };
    };

    // Iterate through timeSeriesResponses accumulating any errors and then
    // broadcast event with them.
    var handleErrors = createMessageHandler('errors', 'chartErrorsChanged');

    // Iterate through timeSeriesResponses accumulating any warnings and then
    // broadcase event with them.
    var handleWarnings = createMessageHandler('warnings', 'chartWarningsChanged');

    self.processPlotContainers = function(timeRange, plotContainers) {
      // Build a map of query -> plotContainer. We use this map later when
      // getting tsdata back from the server to send the appropriate response
      // back to the right plotContainer.
      var queryToPlotContainer = {};
      _.each(plotContainers, function(plotContainer) {
        queryToPlotContainer[plotContainer.getBoundTsquery()] = plotContainer;
      });

      var queryParams = {
        startTime: timeRange.startDate.getTime(),
        endTime: timeRange.endDate.getTime(),
        updateRecent: options.updateRecent || false,
        tsquery: JSON.stringify(_.keys(queryToPlotContainer))
      };

      var timeSeriesUri = "/cmf/charts/timeSeries";
      // TODO: Remove this after we settle on a single backend.
      // Hard code here for now.
      // In Amon, the first character of the first query in the list is {.
      // This must be position 2.
      if (queryParams.tsquery.indexOf("{") === 2) {
        timeSeriesUri = "/cmf/amon/charts/timeSeries";
      }

      return $.post(timeSeriesUri, queryParams, function(response) {
        var data = Util.filterJsonResponseError(response);
        if (!data || !data.length) {
          // TODO: Don't just return?
          return;
        }
        // Handle any errors or warnings if enabled.
        if (enableFeedbackErrors) {
          handleErrors(data);
        }
        if (enableFeedbackWarnings) {
          handleWarnings(data);
        }
        // Iterate through our plot containers and match them with the
        // right data.
        _.each(data, function(timeSeriesResponse) {
          var query = timeSeriesResponse.tsquery;
          var plotContainer = queryToPlotContainer[query];
          if (!plotContainer) {
            return;
          }
          padTsResponse(timeSeriesResponse, queryParams);
          plotContainer.render(timeSeriesResponse);
        });
      });
    };

    // Render each plotContainer. Each plotContainer's associated TsQuery is
    // send to the server asynchronously and its results rendered as soon as
    // the server responds.
    self.render = function(timeRange, preRenderHook, postRenderHook) {
      if (_.isFunction(preRenderHook)) {
        preRenderHook();
      }

      // Save our scroll position.
      var scrollTops = Util.saveScrollTops($container);
      // Iterate over plot containers, sequentially querying server.
      var boundProcessMethod = _.bind(self.processPlotContainers, self, timeRange);
      var bucketedPlotContainers = Util.bucket(options.plotContainers, BUCKET_SIZE);
      var deferreds = _.map(bucketedPlotContainers, boundProcessMethod);

      // Once all the AJAX requests are done, call our postRenderHook and
      // scroll us back where we were.
      // This is annoying: $.when doesn't handle an array parameter the way
      // one would expect, so we have to apply it like this. Bummer. If we
      // don't, then it acts as though all the deferreds are not actually
      // Deferred objects, but are instead resolved promises.
      $.when.apply(null, deferreds).then(function() {
        if (_.isFunction(postRenderHook)) {
          postRenderHook();
        }
        Util.applyScrollTops($container, scrollTops);
      });
    };
  }
  return BulkTsqueryFetcher;
});
